/*
 * $Id: config.cc,v 1.3 2005/05/08 16:37:58 amaia Exp $
 */
#include <iostream>
#include <fstream>
#include <iterator>

#include <boost/program_options.hpp>
#include "version.h"
#include "modular.h"
#include "config.h"

namespace po = boost::program_options;
//using namespace std;
using std::string;
using std::vector;
using std::ostream;
using std::ostream_iterator;
using std::ifstream;
using std::endl;
using std::cout;
using std::cerr;


// A helper function to simplify the main part.
template<class T>
ostream& operator<<(ostream& os, const vector<T>& v)
{
    copy(v.begin(), v.end(), ostream_iterator<T>(cout, " ")); 
    return os;
}


int config::read() 
{
		assert(_argv != 0);
		//group of opts allowd on cmdline only
		po::options_description generic("Generic options");
		generic.add_options()
			("version,v", "print version string")
			("help", "produce help message")
			("config,c", po::value<string>(&cfg_file)->default_value("modular.cfg"), "parameters file")
			;

		//group of opts allowd both on cmdline and config file
		po::options_description cfg("Simulation Parameters");
		cfg.add_options()
			("N,N", po::value<u64>(&N)->default_value(0), "valor N")
			("M,M", po::value<u64>(&M)->default_value(0), "valor M")
			("tau", po::value<ldouble>(&tau)->default_value(0), "tau")
			("U,U", po::value<ldouble>(&U)->default_value(0), "")
			("VEZ,V", po::value<u64>(&VEZ)->default_value(0), "")
			("S2", po::value<ldouble>(&S2)->default_value(0), "")
			("mtarget", po::value<ldouble>(&mtarget)->default_value(0), "")
			("Ttrans", po::value<u64>(&Ttrans)->default_value(0), "")
			("T,T", po::value<u64>(&T)->default_value(0), "")
			("Emax,", po::value<ldouble>(&Emax)->default_value(0), "")
			("sigma_gauss_q", po::value<ldouble>(&sigma_gauss_q)->default_value(0), "")
			("sigma_gauss_p", po::value<ldouble>(&sigma_gauss_p)->default_value(0), "")
			;

		po::options_description cfg2("Configuration Options");
		cfg2.add_options()
			("output_dir,o", po::value<string>(&out_dir)->default_value("out"), "diretorio de saida")
			("input_dir,i", po::value<string>(&inp_dir)->default_value("input"), "diretorio de entrada")
			("debug_level,d", po::value<int>(&debug_level)->default_value(0), "")
			;

/*
		po::options_description hidden("");
		hidden.add_options()
			("input_file,i", po::value<  vector<string> >(&input), "arquivo(s) de entrada")
			;
*/

		po::options_description cmdline_options;
		cmdline_options.add(generic).add(cfg).add(cfg2);//.add(hidden);

		po::options_description cfgfile_options;
		cfgfile_options.add(cfg).add(cfg2);

		po::options_description visible;
		visible.add(generic).add(cfg).add(cfg2);

		po::positional_options_description p;
		p.add("input_file", -1);

		po::variables_map vm;
        store(po::command_line_parser(_argc,_argv).
              options(cmdline_options).positional(p).run(), vm);

		notify(vm);
		
		if(vm.count("config")) {
			ifstream ifs(cfg_file.c_str());
			if(!ifs) {
				cerr << "cant open specified config file, aborting." << endl;
				CkExit();
			}
			store(parse_config_file(ifs, cfgfile_options), vm);
		}

		notify(vm);
   
        if (vm.count("help")) {
			cout << "\n\nModular:\n";
			cout << "Usage: <options> <input-files>" << "\n";
            cout << visible << "\n";
            CkExit();
        }

        if (vm.count("version")) {
            cout << "version: " << current_version << "\n";
            CkExit();
        }

		/*
        if (vm.count("input_file"))
        {
            cout << "Input files are: " 
                 << vm["input_file"].as< vector<string> >() << "\n";
        }
		*/

}

void config::pup(PUP::er &p)
{
	p|N;
	p|M;
	p|tau;
	p|U;
	p|VEZ;
	p|S2;
	p|mtarget;
	p|Ttrans;
	p|T;
	p|Emax;
	p|sigma_gauss_p;
	p|sigma_gauss_q;
	p|cfg_file;
	p|out_dir;
	p|inp_dir;
	p|debug_level;

	//nao preciso fazer para argc/argv
}
   
