/*
 * $Id: modular.cc,v 1.8 2005/05/08 17:47:25 amaia Exp $
 */
#include "modular.h"
using std::vector;
using namespace std;

CProxy_Main mainProxy;
CProxy_simulacao simulProxy;
CkChareID mainhandle;
config gConfig;

int debug_verbosity;


Main::Main(CkArgMsg *m)
: simul_done_count(0)
{
	timer1.start();

	gConfig.init(m->argc,m->argv);
	gConfig.read();

	N 		= gConfig.get_N();
	M 		= gConfig.get_M();
	T 		= gConfig.get_T();
	Ttrans	= gConfig.get_Ttrans();
	tau 	= gConfig.get_tau();
	U		= gConfig.get_U();
	VEZ		= gConfig.get_VEZ();

	DE = 1. / gConfig.get_S2();
	Edim = static_cast<int>(gConfig.get_Emax()/DE);
	
	P_energy_histo	.resize(Edim);
	P_K_energy_histo.resize(Edim);
	P_energy	.resize(Edim);
	P_K_energy	.resize(Edim);
	PP_energy	.resize(Edim);
	real_index	.resize(Edim);
	real_K_index.resize(Edim);
	E_index		.resize(Edim);

	EK_med	.resize(T);
	VM_med	.resize(T);
	Vi_med	.resize(T);
	m_med	.resize(T);
	V		.resize(T);
	K		.resize(T);

	debug_verbosity = gConfig.get_debug_level();

	debug_print(DEBUG_INFO3,"N: %lld ",N);
	debug_print(DEBUG_INFO3,"M: %lld ",M);
	debug_print(DEBUG_INFO3,"Ttrans: %lld ",Ttrans);
	debug_print(DEBUG_INFO3,"T: %lld ",T);
	debug_print(DEBUG_INFO3,"VEZ: %lld ",VEZ);
	debug_print(DEBUG_INFO3,"\n");
	debug_print(DEBUG_INFO3,"tau: %Lf ",tau);
	debug_print(DEBUG_INFO3,"U: %Lf ",U);
	debug_print(DEBUG_INFO3,"S2: %Lf ",gConfig.get_S2());
	debug_print(DEBUG_INFO3,"mtarget: %Lf ",gConfig.get_mtarget());
	debug_print(DEBUG_INFO3,"\n");
	debug_print(DEBUG_INFO3,"Emax: %Lf ", gConfig.get_Emax());
	debug_print(DEBUG_INFO3,"sigma_gauss_q: %Lf ", gConfig.get_sigma_gauss_q());
	debug_print(DEBUG_INFO3,"sigma_gauss_p: %Lf ", gConfig.get_sigma_gauss_p());
	debug_print(DEBUG_INFO3,"\n\n");

	//debug_print(DEBUG_INFO2,"Main::Main()\n");
	
	mainProxy = thishandle;
	mainhandle = thishandle;

	CkStartQD(CkIndex_Main::quiescence(),&mainhandle);

	simulArray = CProxy_simulacao::ckNew(VEZ);
	simulProxy = simulArray;

	//debug_print(DEBUG_INFO2,"\tcalling simulArray.start()\n");
	simulArray.start();
	
}

void Main::quiescence()
{
	timer1.stop();
	debug_print(DEBUG_INFO3,"finished (%f secs).\n",timer1.get());
	CkExit();
}
//	Callback called when a 'simulation' object finishes.
//	The mainchare receives the historical variables,
//	computes occup_index, dynamical averages, histo,
//	and print.
void Main::cbSimulDone(
	const vdouble EK
	,const vdouble VM
	,const vdouble Vi
	,const vdouble m
	,const vdouble v_V
	,const vdouble v_K)
{
	assert( EK.size() == T );
	assert( VM.size() == T );
	assert( Vi.size() == T );
	assert( m.size() == T );
	assert( v_V.size() == T );
	assert( v_K.size() == T );

	vint occup_index		(Edim),
		occup_index_histo	(Edim),
		occup_K_index		(Edim),
		occup_K_index_histo	(Edim);

	timer t1;
	//debug_print(DEBUG_INFO2,"Main::cbSimulDone() - count=%d\n",simul_done_count);
	//debug_print(DEBUG_INFO2,"\t"); 
	//debug_print_ts(DEBUG_INFO2);	

	t1.start();

	//make_occup_index()
	for(u64 i = 0; i < T; i++) {
		int U_cell = static_cast<int>(floor( (EK[i]+VM[i]+DE/2)/DE));
		int K_cell = static_cast<int>(floor((EK[i]+DE/2)/DE));

		assert(U_cell < Edim);

		occup_index[U_cell]++;
		occup_index_histo[U_cell]++;
		occup_K_index[K_cell]++;
		occup_K_index_histo[K_cell]++;
	
	}

	//debug_print(DEBUG_INFO2, "\tmake_occup_index():\n");
	for(int i=0; i<Edim; i++) {
		debug_print(DEBUG_EXTRA, "\toccup_i[%d]=%d, occup_i_h[%d]=%d, occup_k_i[%d]=%d, occup_k_i_h[%d]=%d\n"
			,i,occup_index[i]
			,i,occup_index_histo[i]
			,i,occup_K_index[i]
			,i,occup_K_index_histo[i]);
	}


	//make_E_index()
	for(int i=0; i<Edim; i++) E_index[i] = DE*i;

	update_histo_E(occup_index,occup_index_histo);
	update_histo_K(occup_K_index,occup_K_index_histo);
	
	V = v_V;
	K = v_K;
	
	update_dynamical_averages(EK,VM,Vi,m);
	print_dynamical_averages();
	
	t1.stop();
	debug_print(DEBUG_INFO2,"main::simuldone() - updated stats (%f secs)\n",t1.get());
	simul_done_count++;

	return;
}

void Main::update_histo_E(vint&  occup_index,vint& occup_index_histo)
{
	
	ldouble stddev_tmp;
	ostringstream filename;
	ios_base::openmode mode = simul_done_count ? ios_base::app : ios_base::out;
	filename << gConfig.get_out_dir() << "/o_E.dat";
	ofstream ofs1(filename.str().c_str(),mode);

//	debug_print(DEBUG_INFO2,"\tMain::update_histo_E\n");
	const int count = simul_done_count+1;

	ofs1 << "run " << count << "/" << gConfig.get_VEZ();
	ofs1 << " T/Ttrans:" << gConfig.get_T() << "/" << gConfig.get_Ttrans() << "\n";
	ofs1 << "#index #E #P_energy #stdv_P_energy #counts #counts_acc #P_energy_histo\n";
	for(int i=0; i<Edim; i++) {
		debug_print(DEBUG_EXTRA, "\t\t occup_index[%d]=%d\n",i,occup_index[i]);
		if(occup_index[i] > 0) {
			const int ridx = ++real_index[i];
			P_energy[i] = 
				((ridx-1)*P_energy[i]+(occup_index[i]/(1.*T)))/(1.*ridx);

			PP_energy[i] = 
				((ridx-1)*PP_energy[i]+(occup_index[i]/(1.*T)))/(1.*ridx);
		}
		debug_print(DEBUG_EXTRA,"\t\t P_energy[%d]=%6.3Lf\n",i,P_energy[i]);
		debug_print(DEBUG_EXTRA,"\t\t PP_energy[%d]=%6.3Lf\n",i,PP_energy[i]);

		debug_print(DEBUG_EXTRA, "\t\t occup_index_histo[%d]=%d\n",i,occup_index_histo[i]);
		if(occup_index_histo[i] > 0) {
			P_energy_histo[i] = occup_index_histo[i]/(1.*T*count);
		}

		//write out occupation numbers
		stddev_tmp = PP_energy[i] - P_energy[i]*P_energy[i];
		if(stddev_tmp < 0) stddev_tmp *= -1.;

		if(P_energy_histo[i]) {
			ofs1 << fixed << setprecision(8);
			ofs1 << i << " ";
			ofs1 << E_index[i] << " ";
			ofs1 << P_energy[i] << " ";
			ofs1 << sqrt(stddev_tmp)/2. << " ";
			ofs1 << occup_index[i]/(1.*T) << " ";
			ofs1 << real_index[i] << " ";
			ofs1 << P_energy_histo[i] << "\n";

		}
		
	}

	ofs1.close();
}
void Main::update_histo_K(vint&  occup_index, vint& occup_index_histo)
{
	ldouble stddev_tmp;
//	debug_print(DEBUG_INFO2,"\tMain::update_histo_K\n");
	
	ostringstream filename;
	ios_base::openmode mode = simul_done_count ? ios_base::app : ios_base::out;
	filename << gConfig.get_out_dir() << "/o_K.dat";
	ofstream ofs1(filename.str().c_str(),mode);

	const int count = simul_done_count+1;
	ofs1 << "run " << count << "/" << gConfig.get_VEZ(); 
	ofs1 << " T/Ttrans:" << gConfig.get_T() << "/" << gConfig.get_Ttrans() << "\n";
	ofs1 << "#index #E #P_K_energy #counts #counts_acc #P_energy_histo\n";
	
	for(u64 i=0; i<Edim; i++) {
		if(occup_index[i] > 0) {
			const int ridx = ++real_K_index[i];
			P_K_energy[i] = 
				((ridx-1)*P_K_energy[i]+(occup_index[i]/(1.*T)))/(1.*ridx);
		}

		if(occup_index_histo[i] > 0) {
			P_K_energy_histo[i] = occup_index_histo[i]/(1.*T*count);
		}

		//write out occupation numbers

		if(P_K_energy_histo[i]) {
			ofs1 << fixed << setprecision(8);			
			ofs1 << i << " ";
			ofs1 << E_index[i] << " ";
			ofs1 << P_K_energy[i] << " ";
			ofs1 << occup_index[i]/(1.*T) << " ";
			ofs1 << real_K_index[i] << " ";
			ofs1 << P_K_energy_histo[i] << "\n";

		}
		
	}

	ofs1.close();
}


// Recebe como parametros os valores de EK,VM,Vi,m de uma
// simulacao (todos timesteps) e atualiza as médias
// EK_med, VM_med, Vi_med, m_med.
void Main::update_dynamical_averages(
	const vdouble& EK
	,const vdouble& VM
	,const vdouble& Vi
	,const vdouble& m)
{
//	debug_print(DEBUG_INFO2,"\tMain::update_dynamical_averages\n"); 
	int vez = simul_done_count+1;
	for(u64 i =0; i < T; i++) {
		EK_med[i] = ((vez-1)* EK_med[i]+EK[i])/(1.*vez);
		VM_med[i] = ((vez-1)* VM_med[i]+VM[i])/(1.*vez);
		Vi_med[i] = ((vez-1)* Vi_med[i]+Vi[i])/(1.*vez);
		 m_med[i] = ((vez-1)*  m_med[i]+ m[i])/(1.*vez);
	}

}


void Main::print_dynamical_averages()
{
//	debug_print(DEBUG_INFO2,"\tMain::print_dynamical_averages\n"); 
	//KM+ERR-Uxx-Nxx-Mxx

	ostringstream filename;
	ios_base::openmode mode = simul_done_count ? ios_base::app : ios_base::out;
	filename << gConfig.get_out_dir() << "/km+err.dat";
	ofstream ofs(filename.str().c_str(),mode);

	assert(N == gConfig.get_N());
	assert(U == gConfig.get_U());

	ofs << "vez=" << simul_done_count << "\n";
	for(u64 i =0; i < T; i++) {
		ofs << fixed << setprecision(15);
		ofs << i << " ";
		ofs << 2. * EK_med[i] / M << " ";
		ofs << (EK_med[i]+VM_med[i])/M << " ";
		ofs << m_med[i] << " ";
		ofs << Vi_med[i] << " ";
		ofs <<  100*fabsl(U - (K[i]+V[i])/(1.*N))/U << " ";
		ofs << "\n";
	}
	ofs.close();
}


#include "modular.def.h"
