/*
 * $Id: modular.h,v 1.6 2005/05/08 17:47:25 amaia Exp $
 */
#ifndef __MODULAR_H
#define __MODULAR_H

#include <unistd.h>
#include <sys/time.h>
#include <string>
#include <vector>
#include <iomanip>
#include <cstdarg>
#include <cstdio>
#include <fstream>
#include <sstream>
#include "pup_stl.h"


typedef u_int64_t u64;
typedef long double ldouble;
//typedef double ldouble;
typedef std::vector<ldouble> vdouble;
typedef std::vector<ldouble>::iterator vdouble_it;
typedef std::vector<int> vint;
typedef std::vector<int>::iterator vint_it;

#include "modular.decl.h"
#include "config.h"

enum {
DEBUG_NONE=0,
DEBUG_FATAL=1,
DEBUG_WARN=2,
DEBUG_INFO3=3,
DEBUG_INFO2=4,
DEBUG_INFO1=5,
DEBUG_EXTRA=6
};


int load_pq(const std::string filename, vdouble& p, vdouble& q);
int save_pq(const std::string filename, vdouble& p, vdouble& q);
void debug_print_ts(int level);
void debug_print(int level, const char *fmt, ...);
CkReductionMsg *custom_sum_ldouble(int nMsg, CkReductionMsg **msgs);
CkReduction::reducerType sum_ldouble();
void register_sum_ldouble(void);

class timer {
	struct timeval tv;
	struct timeval tv2;
	double delta;

	int running;
	public:
		timer() :delta(0), running(true) { 
			gettimeofday(&tv, NULL); 
		}
		void start() {
			gettimeofday(&tv, NULL);
			running=true;
		}
		void stop() {
			gettimeofday(&tv2, NULL);
			running=false;
		}
		double get() {
			if(running==true) return -1;
			delta += (tv2.tv_sec - tv.tv_sec);
			delta += ((tv2.tv_usec - tv.tv_usec)/1.e6);
			return delta;
		}
};


class Main : public CBase_Main {
	u64 N;
	u64 M;
	u64 T;
	u64 Ttrans;
	u64 VEZ;
	ldouble tau;
	ldouble U;
	ldouble DE;
	int Edim;

	CProxy_simulacao simulArray;
	int simul_done_count;

	vdouble P_energy 
		,PP_energy 
		,P_energy_histo
		,P_K_energy
		,P_K_energy_histo
		,E_index;

	vint real_index
		,real_K_index;

	vdouble EK_med,
		VM_med,
		Vi_med,
		m_med,
		V,
		K;

	timer timer1;

	private:
	void update_dynamical_averages(
		const vdouble& EK
		,const vdouble& VM
		,const vdouble& Vi
		,const vdouble& m);

	void print_dynamical_averages();
	//averages_and_print()
	void update_histo_E(vint& occup_index, vint& occup_index_histo);
	void update_histo_K(vint& occup_index, vint& occup_index_histo);
	
	public:
	Main(CkArgMsg *m);
	void quiescence();
	void done() { CkExit(); }
	void cbSimulDone(
		const vdouble EK
		,const vdouble VM
		,const vdouble Vi
		,const vdouble m
		,const vdouble V
		,const vdouble K);

};

class rotor : public CBase_rotor
{
	//CProxy_simulacao simulProxy;
	int simulIdx;

	//consts, mas devido ao pup nao posso?
	u64  _N, _M;
	u64 _T, _Ttrans;
	ldouble _tau;

	//tamanho dos vetores p,q
	//qtdade de rotores por objeto
	int _G; 

	//numero efetivo de rotores para o obj 
	//(caso a N % G != 0, sera diferente
	//de G)
	int _arrsize;

	u64 _t;
	int _step;
	ldouble _c[4];
	ldouble _d[4];
	
	vdouble _p,_q;	
	ldouble _K, _KM;

	void init_consts();

	public:
	rotor(CkMigrateMessage *m);
	rotor(int simulIdx, int G);

	void set(vdouble p, vdouble q);

	void calc();
	void calc_stage1();
	void calc_stage2(CkReductionMsg* m);
	
	void pup(PUP::er &p);
	
	~rotor() {}

};

class simulacao : public CBase_simulacao {
	//as classes com pup nao podem possuir atribs const?
	u64 _N, _M, _Ttrans, _T;
	ldouble _tau;
	//ldouble _V, _VM, _Vi, _K, _KM;
	int _G;

	//timestep atual
	int _t;

	CProxy_rotor rotorArray;

	timer timer1;

	//historico dessas variaveis para todos os t
	//dessa simulacao. sera usado posteriormente
	//para calcular 'dynamical averages'(EK_med,VM_med,
	//Vi_med, m_med).
	//
	//(a outra opcao seria utilizar os valores a medida
	//que sao calculados. talvez seja interessante
	//futuramente).
	vdouble v_EK, v_VM, v_Vi, v_m, v_V, v_K;

	public:
	simulacao(CkMigrateMessage* m);
	simulacao();
	void start();
	void cbDone1(CkReductionMsg* m);
	void cbDone2(CkReductionMsg* m);

	void pup(PUP::er &p);

	private:
	void init_rotors(std::string input);
};


#endif
