/*
 * $Id: rotor.cc,v 1.2 2005/05/08 17:47:25 amaia Exp $
 */

#include "modular.h"

using namespace std;


extern CProxy_simulacao simulProxy;
extern config gConfig;
extern int debug_verbosity;

template<class T>
void PUParray(PUP::er& p, T* t, int nelem)
{
	for(int i=0; i<nelem; i++) p|t[i];
}


rotor::rotor(CkMigrateMessage *m)
{ 
	//Este construtor precisa ser criado
	//para dar suporte ao recurso de migracao.
	//a classe preicsa possuir rotina PUP tambem
}

rotor::rotor(int simulIdx, int G)
	:
	simulIdx(simulIdx),
	_G(G),
	_N(gConfig.get_N()),
	_M(gConfig.get_M()),
	_T(gConfig.get_T()),
	_Ttrans(gConfig.get_Ttrans()),
	_tau(gConfig.get_tau()),
	_p(_G),
	_q(_G),
	_step(0)
{ 
	debug_print(DEBUG_EXTRA,"\t\trotor::rotor(simulIdx=%d, G=%d)\n",simulIdx,G);
	init_consts();

	const int nelem = (int)(_N/_G) + (_N % _G == 0 ? 0 : 1);
	// o ultimo obj possui menos que G rotores
	_arrsize = (thisIndex == nelem -1) ? _N - thisIndex*_G : _G;
	
}


//	inicializacao das constantes de integracao
//
void rotor::init_consts()
{
  /*Integration Constants*/
  _c[0] = 1.0/(ldouble)(2.0*(2.0-powl(2.0,1.0/3.0)));
  _c[1] = (1.0-powl(2.0,1.0/3.0))*_c[0];
  _c[2] = _c[1];
  _c[3] = _c[0];

  _d[0] = 1.0/(ldouble)(2.0-powl(2.0,1.0/3.0));
  _d[1] = -powl(2.0,1.0/3.0)*_d[0];
  _d[2] = _d[0];
  _d[3] = 0.;

#if 0
  debug_print(DEBUG_EXTRA,"Constantes de integracao:\n");
  debug_print(DEBUG_EXTRA,"c: %Lf %Lf %Lf %Lf\n",_c[0],_c[1],_c[2],_c[3]);
  debug_print(DEBUG_EXTRA,"d: %Lf %Lf %Lf %Lf\n",_d[0],_d[1],_d[2],_d[3]);
#endif

}


void rotor::set(vdouble p, vdouble q)

{
	assert(p.size() == q.size());
	assert(p.size() == _G);
	assert(p.size() == _arrsize);

	_p.clear();
	_q.clear();
	
	_p.reserve(_G);
	_q.reserve(_G);

	copy( p.begin(), p.end(), back_inserter<vdouble>(_p));
	copy( q.begin(), q.end(), back_inserter<vdouble>(_q));

	assert(_p.size() == _arrsize);
	assert(_q.size() == _arrsize);
}

void rotor::calc()
{	
	_t		= 0;
	_step	= 0;
	calc_stage1();
}


//
//	Estágio 1:
//		- calcula q a partir do p
//		- calcula as componentes de m devido a esse q
//		- contribui esses valores de m para o reduction
//
void rotor::calc_stage1()
{
	debug_print(DEBUG_INFO1,"stage1 - start - obj %d , t %d step %d\n",thisIndex,_t,_step);
	
	ldouble mx=0, my=0, mx_M=0, my_M=0;

	//base indica o indice real do primeiro rotor deste objeto
	const ldouble base = thisIndex * _G;

	assert(_p.size() == _arrsize);
	for(int i=0; i< _arrsize; i++) {
		
		_q[i]	+= _c[_step] * _p[i] * _tau;

		while(_q[i] < 0) _q[i] += 2. * M_PI;
		_q[i]	= fmod(_q[i] ,2. * M_PI);

		mx 		+= cosl(_q[i]);
		my 		+= sinl(_q[i]);
		mx_M 	+= (i+ base < _M) ? cosl(_q[i]) : 0;
		my_M 	+= (i+ base < _M) ? sinl(_q[i]) : 0;
	
	}

	mx   /=_N;
	my   /=_N;
	mx_M /=_M;
	my_M /=_M;

#if 0
	ostringstream filename;
	filename << "pq-" << _step << "-" << 1;
	save_pq(filename.str(),_p,_q);
#endif

//	debug_print(DEBUG_INFO1,"\tstage1 - end   - obj %d (%Lf,%Lf,%Lf,%Lf)\n",thisIndex,mx,mx_M,my,my_M);
	CkCallback cb(CkIndex_rotor::calc_stage2(NULL),thisProxy);
	ldouble res[] = { mx, my, mx_M, my_M };
	contribute(sizeof(res),res,sum_ldouble(),cb);
	
	return;
}


//	Estágio 2:
//		-recebe as somas das componentes m
//		-calcula q a partir desses m (e F)
//
void rotor::calc_stage2(CkReductionMsg* m)
{
//	debug_print(DEBUG_INFO1,"stage2 - start - obj %d , t %d step %d\n",thisIndex,_t,_step);
	const ldouble base = thisIndex * _G;
	
	ldouble mx,my,mx_M,my_M;
	ldouble *res = (ldouble*) m->getData();
	mx  	= *res++;
	my 		= *res++;
	mx_M	= *res++;
	my_M	= *res++;
	delete m;

//	debug_print(DEBUG_EXTRA,"\t(%Lf,%Lf,%Lf,%Lf)\n",mx,mx_M,my,my_M);
//	debug_print("t=%d step=%d mx=%f my=%f mx_M=%f my_M=%f\n",_t,_step,mx,my,mx_M,my_M);
	assert(_p.size() == _arrsize);

	_K = .0;
	_KM = .0;
	
	for(int i=0; i< _arrsize; i++) {
	
		const ldouble F = -mx * sinl(_q[i]) + my * cosl(_q[i]);
		_p[i]	+= _d[_step] * F * _tau;
		
		if(_step == 3) {
			//fourth step, calculation of p[i] is finished
			const ldouble k = .5 * _p[i] * _p[i];
			_K += k;
			_KM += (i+base<_M)? k : 0;
		}
	}

	ldouble V=.0 ,VM=.0 ,Vi=.0, m_temp=.0 ;
	if(_step == 3 && thisIndex == 0) {
		V  = .5 * (1 - mx*mx - my*my) * _N;
		VM = .5 * _M * (1 - mx_M*mx_M - my_M*my_M);
		Vi = _M * (_N - _M) - _M * ( mx_M * (mx*_N - mx_M*_M) + my_M * (my*_N - my_M*_M) );
		Vi/= _M * (_N - _M - 1);
		
		m_temp = mx_M*mx_M + my_M*my_M;
		
	}


//	debug_print(DEBUG_INFO1,"stage2 - end   - obj %d \n",thisIndex);

#if 0
	ostringstream filename;
	filename << "pq-" << _step << "-" << 2;
	save_pq(filename.str(),_p,_q);
#endif

	++_step;

	//cada iteracao possui 4 steps
	if(_step < 4) {
	
//		debug_print("rotor::done_stage2() step=%d done\n",_step);
		calc_stage1();
		
	} else {

		//estou fazendo reduction com K, KM, apesar de esse valor
		//so importar no ultimo step (=3). melhorar isso..
//		debug_print(DEBUG_INFO1,"\t contrib 1 index=%d\n",thisIndex);
		CkCallback cb1(CkIndex_simulacao::cbDone1(NULL),CkArrayIndex1D(simulIdx),simulProxy);
		ldouble res1[] = { V, VM, Vi, _K, _KM, m_temp};
		
//		debug_print(DEBUG_INFO1,"\t contribute(6 *sizeof(ldouble))=(6 * %d)=%d\n",sizeof(ldouble),sizeof(res1));
		contribute(sizeof(res1),res1,sum_ldouble(),cb1);
	
#if 0
		//obs: o codigo abaixo é para uma operacao de reduction
		//que passara para o obj simulacao os estados(p,q) de todos
		//os rotores. nas proximas versoes, esse reduction sera desnecessario,
		CkCallback cb2(CkIndex_simulacao::cbDone2(NULL), simulProxy);
		//	na contribuição, o thisIndex identifica o indice 
		//	do rotor assim o callback terá como ordenar 
		//	os valores por esse índice.
		//double res[] = { thisIndex }; //, _p, _q };
		//res[] = { thisIndex, K, KM, p[0], q[0], p[1], q[1], ... }
		ldouble res2[2*arrsize+1];
		res2[0]=thisIndex;
		//debug_print("contributign index=%d\n",thisIndex);
		for(int i=0; i<2*arrsize; i+=2) {
			res2[i+1] = _p[i/2];
			res2[i+2] = _q[i/2];
		}
		debug_print(DEBUG_INFO1,"\t contrib 2 idx=%d (%d +%d elem)\n",thisIndex,1,-1+sizeof(res2)/sizeof(res2[0]));
		debug_print(DEBUG_INFO1,"\t contribute( (2*arrsize+1)=%d * sizeof(ldouble))\n",1+2*arrsize);
		contribute(sizeof(res2),res2,CkReduction::set,cb2);
#endif

		//terminada essa iteração, o objeto continuara
		//fazendo t=t+1
		_step = 0;
		_t++;
		if(_t < _T) {
			calc_stage1();
		}
		//senao, fim da simulação
	}	


	return;
}


void rotor::pup(PUP::er &p)
{
	ArrayElement1D::pup(p);
	p|_G;
	p|_N;
	p|_M;
	p|_T;
	p|_Ttrans;
	p|_tau;
	p|_t;
	p|_step;
	p|_p;
	p|_q;
	PUParray(p,_c,4);
	PUParray(p,_d,4);

}

