/*
 * $Id: simulacao.cc,v 1.2 2005/05/08 17:47:25 amaia Exp $
 */

#include "modular.h"

using namespace std;

extern CProxy_Main mainProxy;
extern CkChareID mainhandle;
extern config gConfig;
extern int debug_verbosity;

///
//	inicializa os vetores p, q com N elementos a partir do
//  arquivo de entrada
void initial_mom_coor(string filename, vdouble& p, vdouble& q, u64 N)
{
	timer t1;
	t1.start();
	
//	debug_print(DEBUG_INFO2,"init_mom_coor() - start\n"); 

	p.clear();
	q.clear();
	p.reserve(N);
	q.reserve(N);
	if(load_pq(filename,p,q) < 0) {
		debug_print(DEBUG_FATAL,"Error reading file (%s).\n",filename.c_str());
		CkAbort("");
	}
	
	t1.stop();
	//debug_print(DEBUG_INFO2,"init_mom_coor() (t= %f secs)\n",t1.get());
}


//
//
//
//	Classe simulacao
//
//
//

simulacao::simulacao(CkMigrateMessage *m)
{ 
	//este construtor dara suporte ao recurso
	//de migracao do charm. escrever tambem
	//a rotina de pup para esta classe
}



simulacao::simulacao()
	:
	_N(gConfig.get_N()),
	_M(gConfig.get_M()),
	_T(gConfig.get_T()),
	_Ttrans(gConfig.get_Ttrans()),
	_tau(gConfig.get_tau()),

	//por default, setar G como sendo N
	_G(gConfig.get_N()),
	_t(0)
{ 
	v_EK.reserve(_T);
	v_VM.reserve(_T);
	v_Vi.reserve(_T);
	v_m.reserve(_T);
	v_V.reserve(_T);
	v_K.reserve(_T);

	
	//a1=CProxy_A1::ckNew(parameters,nElements); )
	const int nelem = (int)(_N/_G) + (_N % _G == 0 ? 0 : 1);
	debug_print(DEBUG_INFO2,"simulacao::simulacao() creating %d objects\n",nelem);
	
	rotorArray = CProxy_rotor::ckNew(thisIndex,_G,nelem);
	init_rotors(gConfig.get_inp_dir());

}

//	
//	inicializa o array de rotores com os vetores p,q
//
void simulacao::init_rotors(string input)
{
	
//	debug_print(DEBUG_INFO2,"simul::init_rotors()\n");
	const int nelem = (int)(_N/_G) + (_N % _G == 0 ? 0 : 1);

	//inicializar p e q
	vdouble p, q;
	vdouble_it pit, qit;

	ostringstream filename;
	filename << input << "/input." << thisIndex;

//	debug_print(DEBUG_INFO2, "simul::init_rotors(), opening %s\n",filename.str().c_str());
	initial_mom_coor(filename.str(),p,q,_N);
	assert(p.size() == q.size());
	
	pit = p.begin();
	qit = q.begin();
	int n;
	for(int i=0; i<nelem; i++) {
		// n sera ou _G ou a qtdade de elems q faltam
		n = (i == nelem -1) ? _N - i*_G : _G;
		
		rotorArray[i].set ( vdouble(pit,pit+n), vdouble(qit,qit+n) );
		//debug_print("simul::init_rotors() - setting rotor[%d] n=%d\n",i,n);
		pit += n;
		qit += n;
	}	
//	debug_print(DEBUG_INFO2,"simul::init_rotors()-end\n");
	return;
}

void simulacao::start()
{
	//debug_print(DEBUG_INFO2,"simul::start()\n");
	timer1.start();
	rotorArray.calc();
}

void simulacao::cbDone1(CkReductionMsg* m)
{
	ldouble V, VM, Vi, K, KM;
	ldouble* data = (ldouble*) m->getData();
	assert(m->getSize() == 6*sizeof(ldouble));	
	
	v_V.push_back(*data++);
	v_VM.push_back(*data++);
	v_Vi.push_back(*data++);
	v_K.push_back(*data++);
	v_EK.push_back(*data++);
	v_m.push_back(*data++);
	
	delete m;

//	debug_print(DEBUG_INFO1,"simulacao::cbDone1()\n");
//	debug_print(DEBUG_EXTRA,
//		"\tV=%Lf,VM=%Lf,Vi=%Lf,K=%Lf,EK=%Lf,m=%Lf\n"
//		,v_V[_t],v_VM[_t],v_Vi[_t],v_K[_t],v_EK[_t],v_m[_t]);

	_t++;
	
	if(_t==_T) {
		
		timer1.stop();
		debug_print(DEBUG_INFO2,"\tsimulacao %d done. (%f secs)\n", thisIndex,timer1.get());
		timer1.start();
		
		mainProxy.cbSimulDone(v_EK,v_VM,v_Vi,v_m,v_V,v_K);
	}
}


//
//	Este metodo eh desncessario..
//
void simulacao::cbDone2(CkReductionMsg* m)
{

#if 0
	//Esse metodo eh o Callback do Reduction que reune
	//todos os (p,q) dos rotores no obj simulacao, caso
	//haja necessidade de imprimir (debugging,etc).

	static int x =0;
	vdouble p(_N), q(_N);
	int idx; 

	debug_print(DEBUG_INFO2, "simulacao %d iteração t=%d\n",thisIndex,x);
	
	CkReduction::setElement* cur = (CkReduction::setElement*) m->getData();
	while(cur!=NULL) {

		int size = cur->dataSize/sizeof(ldouble);
		ldouble* data = (ldouble*)cur->data;

		// o primeiro elemento é o índice do
		// rotor que possui esses valores de p e q.
		idx		= *data++; size--;
		debug_print(DEBUG_INFO1,"simulacao::cbDone2() - idx=%d, size=%d \n",idx,size+1);
		assert(idx >= 0 && idx < _N);

		//base=real_index
		int base = idx*_G;
		//debug_print("idx=%d size=%d\n",base,size);
		for(int i=0; i<size; i+=2) {
			p[base] = *data++;
			q[base] = *data++;
			//debug_print("idx=%d p %f q %f\n",base,p[base],q[base]);
			base++;
		}
		
		cur=cur->next();
	}


	
	//nao eh mais necessario, usnado quiescence detection
	//if(x==1000) { mainProxy.done(); }

	ostringstream filename;
	filename << "out-" << thisIndex << "." << x;
	save_pq(filename.str(),p,q);
	x++;
	delete m;

#endif

}

void simulacao::pup(PUP::er &p)
{
	//desatualizado!
	ArrayElement1D::pup(p);
	p|_N;
	p|_M;
	p|_T;
	p|_Ttrans;
	p|_tau;	
	p|rotorArray;

}

